@@ -15,6 +15,7 @@
 
 #include <stdio.h>
 #include <signal.h>
+#include <stdint.h>
 #include "coq_gc.h"
 #include "coq_instruct.h"
 #include "coq_fix_code.h"
@@ -30,9 +31,9 @@
 #endif
 
 /* spiwack: I append here a few macros for value/number manipulation */
-#define uint32_of_value(val) (((uint32)val >> 1))
-#define value_of_uint32(i)   ((value)(((uint32)(i) << 1) | 1))
-#define UI64_of_uint32(lo)  ((uint64)(I64_literal(0,(uint32)(lo))))
+#define uint32_of_value(val) (((uint32_t)val >> 1))
+#define value_of_uint32(i)   ((value)(((uint32_t)(i) << 1) | 1))
+#define UI64_of_uint32(lo)  ((uint64_t)(I64_literal(0,(uint32_t)(lo))))
 #define UI64_of_value(val) (UI64_of_uint32(uint32_of_value(val)))
 /* /spiwack */
 
@@ -76,16 +77,11 @@ sp is a local copy of the global variable extern_sp. */
 #ifdef _COQ_DEBUG_ 
 #   define print_instr(s) /*if (drawinstr)*/ printf("%s\n",s)
 #   define print_int(i)   /*if (drawinstr)*/ printf("%d\n",i)
+#   define print_lint(i)  /*if (drawinstr)*/ printf("%ld\n",i)
 # else 
 #   define print_instr(s) 
 #   define print_int(i) 
-#endif
-
-/* Wrapper pour caml_modify */ 	 
-#ifdef OCAML_307 	 
-#define CAML_MODIFY(a,b) modify(a,b) 	 
-#else 	 
-#define CAML_MODIFY(a,b) caml_modify(a,b) 	 
+#   define print_lint(i)
 #endif
 
 /* GC interface */
@@ -145,7 +141,7 @@ sp is a local copy of the global variable extern_sp. */
 #define SP_REG asm("a4")
 #define ACCU_REG asm("d7")
 #endif
-#ifdef __arm__
+#if defined(__arm__) && !defined(__thumb2__)
 #define PC_REG asm("r9")
 #define SP_REG asm("r8")
 #define ACCU_REG asm("r7")
@@ -158,7 +154,7 @@ sp is a local copy of the global variable extern_sp. */
 #endif
 #endif
 
-/* For signal handling, we highjack some code from the caml runtime */
+/* For signal handling, we hijack some code from the caml runtime */
 
 extern intnat caml_signals_are_pending;
 extern intnat caml_pending_signals[];
@@ -345,6 +341,7 @@ value coq_interprete
       /* Fallthrough */
       Instruct(ENVACC){
 	print_instr("ENVACC");
+	print_int(*pc);
 	accu = Field(coq_env, *pc++);
         Next;
       }
@@ -375,6 +372,10 @@ value coq_interprete
 	sp[1] = (value)pc;
 	sp[2] = coq_env;
 	sp[3] = Val_long(coq_extra_args);
+	print_instr("call stack=");
+	print_lint(sp[1]);
+	print_lint(sp[2]);
+	print_lint(sp[3]);
 	pc = Code_val(accu);
 	coq_env = accu;
 	coq_extra_args = 0;
@@ -462,6 +463,7 @@ value coq_interprete
 	sp[0] = arg1;
 	sp[1] = arg2;
 	pc = Code_val(accu);
+	print_lint(accu);
 	coq_env = accu;
 	coq_extra_args += 1;
 	goto check_stacks;
@@ -485,11 +487,18 @@ value coq_interprete
 	print_instr("RETURN");
 	print_int(*pc);
 	sp += *pc++;
+	print_instr("stack=");
+	print_lint(sp[0]);
+	print_lint(sp[1]);
+	print_lint(sp[2]);
 	if (coq_extra_args > 0) {
+	  print_instr("extra args > 0");
+	  print_lint(coq_extra_args);
 	  coq_extra_args--;
 	  pc = Code_val(accu);
 	  coq_env = accu;
 	} else {
+	  print_instr("extra args = 0");
 	  pc = (code_t)(sp[0]);
 	  coq_env = sp[1];
 	  coq_extra_args = Long_val(sp[2]);
@@ -550,21 +559,21 @@ value coq_interprete
 	    coq_extra_args = Long_val(sp[2]);
 	    sp += 3;
 	  } else {
-	    /* L'argument recursif est un accumulateur */
+	    /* The recursif argument is an accumulator */
 	    mlsize_t num_args, i;
-	    /* Construction du PF partiellement appliqu */ 
+	    /* Construction of partially applied PF */
 	    Alloc_small(accu, rec_pos + 2, Closure_tag); 
 	    Field(accu, 1) = coq_env;
 	    for (i = 0; i < rec_pos; i++) Field(accu, i + 2) = sp[i];
 	    Code_val(accu) = pc;
 	    sp += rec_pos;
 	    *--sp = accu;
-	      /* Construction de l'atom */
+	      /* Construction of the atom */
 	    Alloc_small(accu, 2, ATOM_FIX_TAG);
 	    Field(accu,1) = sp[0];
 	    Field(accu,0)  = sp[1];
 	    sp++; sp[0] = accu;
-	      /* Construction de l'accumulateur */
+	      /* Construction of the accumulator */
 	    num_args = coq_extra_args - rec_pos;
 	    Alloc_small(accu, 2+num_args, Accu_tag);
 	    Code_val(accu) = accumulate;
@@ -589,7 +598,10 @@ value coq_interprete
 	Alloc_small(accu, 1 + nvars, Closure_tag);
 	Code_val(accu) = pc + *pc;
 	pc++;
-	for (i = 0; i < nvars; i++) Field(accu, i + 1) = sp[i];
+	for (i = 0; i < nvars; i++) {
+	  print_lint(sp[i]);
+	  Field(accu, i + 1) = sp[i];
+	}
 	sp += nvars;
 	Next;
       }
@@ -671,7 +683,7 @@ value coq_interprete
 	  Field(accu, 0) = sp[0];
 	  *sp = accu;
 	  /* mise a jour du block accumulate */
-	  CAML_MODIFY(&Field(p[i], 1),*sp);
+	  caml_modify(&Field(p[i], 1),*sp);
 	  sp++;
 	}
 	pc += nfunc;
@@ -724,6 +736,7 @@ value coq_interprete
       /* Fallthrough */
       Instruct(GETGLOBAL){
 	print_instr("GETGLOBAL");
+	print_int(*pc);
 	accu = Field(coq_global_data, *pc);
         pc++;
         Next;
@@ -736,7 +749,7 @@ value coq_interprete
 	tag_t tag = *pc++;
 	mlsize_t i;
 	value block;
-	print_instr("MAKEBLOCK");
+	print_instr("MAKEBLOCK, tag=");
 	Alloc_small(block, wosize, tag);
 	Field(block, 0) = accu;
 	for (i = 1; i < wosize; i++) Field(block, i) = *sp++;
@@ -747,7 +760,8 @@ value coq_interprete
 	
 	tag_t tag = *pc++;
 	value block;
-	print_instr("MAKEBLOCK1");
+	print_instr("MAKEBLOCK1, tag=");
+	print_int(tag);
 	Alloc_small(block, 1, tag);
 	Field(block, 0) = accu;
 	accu = block;
@@ -757,7 +771,8 @@ value coq_interprete
 	
 	tag_t tag = *pc++;
 	value block;
-	print_instr("MAKEBLOCK2");
+	print_instr("MAKEBLOCK2, tag=");
+	print_int(tag);
 	Alloc_small(block, 2, tag);
 	Field(block, 0) = accu;
 	Field(block, 1) = sp[0];
@@ -768,7 +783,8 @@ value coq_interprete
       Instruct(MAKEBLOCK3) {
 	tag_t tag = *pc++;
 	value block;
-	print_instr("MAKEBLOCK3");
+	print_instr("MAKEBLOCK3, tag=");
+	print_int(tag);
 	Alloc_small(block, 3, tag);
 	Field(block, 0) = accu;
 	Field(block, 1) = sp[0];
@@ -780,7 +796,8 @@ value coq_interprete
       Instruct(MAKEBLOCK4) {
 	tag_t tag = *pc++;
 	value block;
-	print_instr("MAKEBLOCK4");
+	print_instr("MAKEBLOCK4, tag=");
+	print_int(tag);
 	Alloc_small(block, 4, tag);
 	Field(block, 0) = accu;
 	Field(block, 1) = sp[0];
@@ -795,18 +812,18 @@ value coq_interprete
 /* Access to components of blocks */
         
       Instruct(SWITCH) {
-	uint32 sizes = *pc++;
+	uint32_t sizes = *pc++;
 	print_instr("SWITCH");
-	print_int(sizes & 0xFFFF);
+	print_int(sizes & 0xFFFFFF);
 	if (Is_block(accu)) {
 	  long index = Tag_val(accu);
 	  print_instr("block");
-	  print_int(index);
-	  pc += pc[(sizes & 0xFFFF) + index];
+	  print_lint(index);
+	  pc += pc[(sizes & 0xFFFFFF) + index];
 	} else {
 	  long index = Long_val(accu);
 	  print_instr("constant");
-	  print_int(index);
+	  print_lint(index);
 	  pc += pc[index];
 	}
 	  Next;
@@ -842,15 +859,14 @@ value coq_interprete
       
       Instruct(SETFIELD0){
 	print_instr("SETFIELD0");
-	CAML_MODIFY(&Field(accu, 0),*sp);
+	caml_modify(&Field(accu, 0),*sp);
 	sp++;
 	Next;
       }
 	
       Instruct(SETFIELD1){
-        int i, j, size, size_aux;
 	print_instr("SETFIELD1");
-	CAML_MODIFY(&Field(accu, 1),*sp);
+	caml_modify(&Field(accu, 1),*sp);
        	sp++;
 	Next; 
       }
@@ -868,9 +884,9 @@ value coq_interprete
 	    *sp = accu;
 	    Alloc_small(accu, 1, ATOM_COFIX_TAG);
 	    Field(accu, 0) = Field(Field(*sp, 1), 0);
-	    CAML_MODIFY(&Field(*sp, 1), accu);
+	    caml_modify(&Field(*sp, 1), accu);
 	    accu = *sp; sp++;
-            CAML_MODIFY(&Field(*sp, i), accu);
+            caml_modify(&Field(*sp, i), accu);
 	  }
 	}
 	sp++;
@@ -879,11 +895,35 @@ value coq_interprete
       
       Instruct(SETFIELD){
 	print_instr("SETFIELD");
-	CAML_MODIFY(&Field(accu, *pc),*sp);
+	caml_modify(&Field(accu, *pc),*sp);
 	sp++; pc++;
 	Next;
-      }	
-   
+      }
+
+
+      Instruct(PROJ){
+	print_instr("PROJ");
+	if (Is_accu (accu)) {
+          value block;
+	  /* Skip over the index of projected field */
+	  pc++;
+	  /* Create atom */
+	  Alloc_small(block, 2, ATOM_PROJ_TAG);
+	  Field(block, 0) = Field(coq_global_data, *pc);
+	  Field(block, 1) = accu;
+          accu = block;
+          /* Create accumulator */
+          Alloc_small(block, 2, Accu_tag);
+          Code_val(block) = accumulate;
+          Field(block, 1) = accu;
+          accu = block;
+	} else {
+	    accu = Field(accu, *pc++);
+	}
+	pc++;
+	Next;
+      }
+
 /* Integer constants */
 
       Instruct(CONST0){
@@ -923,32 +963,13 @@ value coq_interprete
       /* Fallthrough */
       Instruct(CONSTINT) {
 	print_instr("CONSTINT");
+	print_int(*pc);
 	accu = Val_int(*pc);
 	pc++;
 	Next;
       }
 
       /* Special operations for reduction of open term */
-      Instruct(ACCUMULATECOND) {
-	int i, num;
-	print_instr("ACCUMULATECOND");
-	num = *pc;
-	pc++;
-	if (Field(coq_global_boxed, num) == Val_false || coq_all_transp) {
-	  /*	  printf ("false\n");
-		  printf ("tag = %d", Tag_val(Field(accu,1))); */
-	  num = Wosize_val(coq_env);
-	  for(i = 2; i < num; i++) *--sp = Field(accu,i);
-	  coq_extra_args = coq_extra_args + (num - 2);
-	  coq_env = Field(Field(accu,1),1);
-	  pc = Code_val(coq_env);
-	  accu = coq_env;
-	  /*	  printf ("end\n"); */
-	  Next;
-	};
-	/*	printf ("true\n"); */
-      }
-      
       Instruct(ACCUMULATE) {
 	mlsize_t i, size;
 	print_instr("ACCUMULATE");
@@ -983,7 +1004,7 @@ value coq_interprete
 	    sp -= nargs;
 	    for (i = 0; i < nargs; i++) sp[i] = Field(accu, i + 2);
 	    *--sp = accu;
-	    print_int(nargs);
+	    print_lint(nargs);
 	    coq_extra_args = nargs;
 	    pc = Code_val(coq_env);
 	    goto check_stacks;
@@ -1081,7 +1102,7 @@ value coq_interprete
            the one ontop of the stack (which is poped)*/
         print_instr("ADDINT31");
         accu = 
-           (value)((uint32) accu + (uint32) *sp++ - 1); 
+           (value)((uint32_t) accu + (uint32_t) *sp++ - 1); 
         /* nota,unlike CaML we don't want
            to have a different behavior depending on the 
            architecture. Thus we cast the operand to uint32 */
@@ -1091,9 +1112,9 @@ value coq_interprete
       Instruct (ADDCINT31) {
 	print_instr("ADDCINT31");
 	/* returns the sum with a carry */
-	uint32 s;
-	s = (uint32)accu + (uint32)*sp++ - 1;
-        if( (uint32)s < (uint32)accu ) {
+	uint32_t s;
+	s = (uint32_t)accu + (uint32_t)*sp++ - 1;
+        if( (uint32_t)s < (uint32_t)accu ) {
           /* carry */
 	  Alloc_small(accu, 1, 2); /* ( _ , arity, tag ) */
 	}
@@ -1108,10 +1129,9 @@ value coq_interprete
       Instruct (ADDCARRYCINT31) {
 	print_instr("ADDCARRYCINT31");
 	/* returns the sum plus one with a carry */
-	uint32 s;
-	s = (uint32)accu + (uint32)*sp++ + 1;
-	value block;
-        if( (uint32)s <= (uint32)accu ) {
+	uint32_t s;
+	s = (uint32_t)accu + (uint32_t)*sp++ + 1;
+        if( (uint32_t)s <= (uint32_t)accu ) {
           /* carry */
 	  Alloc_small(accu, 1, 2); /* ( _ , arity, tag ) */
 	}
@@ -1127,18 +1147,18 @@ value coq_interprete
 	print_instr("SUBINT31");
 	/* returns the subtraction */
 	accu = 
-	  (value)((uint32) accu - (uint32) *sp++ + 1); 
+	  (value)((uint32_t) accu - (uint32_t) *sp++ + 1); 
         Next;
       }
 
       Instruct (SUBCINT31) {
 	print_instr("SUBCINT31");
 	/* returns the subtraction with a carry */
-	uint32 b;
-	uint32 s;
-	b = (uint32)*sp++;
-	s = (uint32)accu - b + 1;
-        if( (uint32)accu < b ) {
+	uint32_t b;
+	uint32_t s;
+	b = (uint32_t)*sp++;
+	s = (uint32_t)accu - b + 1;
+        if( (uint32_t)accu < b ) {
           /* carry */
 	  Alloc_small(accu, 1, 2); /* ( _ , arity, tag ) */
 	}
@@ -1153,11 +1173,11 @@ value coq_interprete
       Instruct (SUBCARRYCINT31) {
 	print_instr("SUBCARRYCINT31");
 	/* returns the subtraction minus one with a carry */
-	uint32 b;
-	uint32 s;
-	b = (uint32)*sp++;
-	s = (value)((uint32)accu - b - 1);
-        if( (uint32)accu <= b ) {
+	uint32_t b;
+	uint32_t s;
+	b = (uint32_t)*sp++;
+	s = (value)((uint32_t)accu - b - 1);
+        if( (uint32_t)accu <= b ) {
           /* carry */
 	  Alloc_small(accu, 1, 2); /* ( _ , arity, tag ) */
 	}
@@ -1181,7 +1201,7 @@ value coq_interprete
         /*returns the multiplication on a double size word
           (special case for 0) */
         print_instr("MULCINT31");
-	uint64 p;
+	uint64_t p;
         /*accu = 2v+1, *sp=2w+1 ==> p = 2v*w */
         p = I64_mul (UI64_of_value (accu), UI64_of_uint32 ((*sp++)^1));
 	if ( I64_is_zero(p) ) {
@@ -1204,10 +1224,10 @@ value coq_interprete
 	/* spiwack: takes three int31 (the two first ones represent an
                     int62) and performs the euclidian division of the
                     int62 by the int31 */
-        uint64 bigint;
+        uint64_t bigint;
         bigint = UI64_of_value(accu);
         bigint = I64_or(I64_lsl(bigint, 31),UI64_of_value(*sp++));
-        uint64 divisor;
+        uint64_t divisor;
         divisor = UI64_of_value(*sp++);
         Alloc_small(accu, 2, 1); /* ( _ , arity, tag ) */
         if (I64_is_zero (divisor)) {
@@ -1215,7 +1235,7 @@ value coq_interprete
           Field(accu, 1) = 1; /* 2*0+1 */
 	}
 	else {
-	  uint64 quo, mod;
+	  uint64_t quo, mod;
           I64_udivmod(bigint, divisor, &quo, &mod);
           Field(accu, 0) = value_of_uint32(I64_to_int32(quo));
 	  Field(accu, 1) = value_of_uint32(I64_to_int32(mod));
@@ -1228,7 +1248,7 @@ value coq_interprete
         /* spiwack: a priori no need of the NON_STANDARD_DIV_MOD flag
                     since it probably only concerns negative number.
                     needs to be checked at this point */
-        uint32 divisor;
+        uint32_t divisor;
         divisor = uint32_of_value(*sp++);
         if (divisor == 0) { 
           Alloc_small(accu, 2, 1); /* ( _ , arity, tag ) */
@@ -1236,7 +1256,7 @@ value coq_interprete
           Field(accu, 1) = 1; /* 2*0+1 */
         }
         else {
-          uint32 modulus;
+          uint32_t modulus;
           modulus = uint32_of_value(accu);
           Alloc_small(accu, 2, 1); /* ( _ , arity, tag ) */
 	  Field(accu, 0) = value_of_uint32(modulus/divisor);
@@ -1248,35 +1268,36 @@ value coq_interprete
       Instruct (ADDMULDIVINT31) {
         print_instr("ADDMULDIVINT31");
         /* higher level shift (does shifts and cycles and such) */
-        uint32 shiftby;
+        uint32_t shiftby;
         shiftby = uint32_of_value(accu);
         if (shiftby > 31) {
           if (shiftby < 62) {
-            *sp++;
-	    accu = (value)((((*sp++)^1) << (shiftby - 31)) | 1);
+            sp++;
+	    accu = (value)(((((uint32_t)*sp++)^1) << (shiftby - 31)) | 1);
           }
           else {
+            sp+=2;
             accu = (value)(1);
           }
 	}
         else{
         /* *sp = 2*x+1 --> accu = 2^(shiftby+1)*x */
-        accu = (value)(((*sp++)^1) << shiftby);
+	  accu = (value)((((uint32_t)*sp++)^1) << shiftby);
         /* accu = 2^(shiftby+1)*x --> 2^(shifby+1)*x+2*y/2^(31-shiftby)+1 */
-        accu = (value)((accu | (((uint32)(*sp++)) >> (31-shiftby)))|1);
+        accu = (value)((accu | (((uint32_t)(*sp++)) >> (31-shiftby)))|1);
 	}
         Next;
       }
 
       Instruct (COMPAREINT31) {
 	/* returns Eq if equal, Lt if accu is less than *sp, Gt otherwise */
-	/* assumes Inudctive _ : _ := Eq | Lt | Gt */
+	/* assumes Inductive _ : _ := Eq | Lt | Gt */
 	print_instr("COMPAREINT31");
-	if ((uint32)accu == (uint32)*sp) {
+	if ((uint32_t)accu == (uint32_t)*sp) {
 	  accu = 1;  /* 2*0+1 */
 	  sp++;
 	}
-	else{if ((uint32)accu < (uint32)(*sp++)) {
+	else{if ((uint32_t)accu < (uint32_t)(*sp++)) {
 	  accu = 3;  /* 2*1+1 */
 	}
 	else{
@@ -1287,9 +1308,9 @@ value coq_interprete
  
       Instruct (HEAD0INT31) {
 	int r = 0;
-        uint32 x;
+        uint32_t x;
         print_instr("HEAD0INT31");
-	x = (uint32) accu;
+	x = (uint32_t) accu;
         if (!(x & 0xFFFF0000)) { x <<= 16; r += 16; }
         if (!(x & 0xFF000000)) { x <<= 8;  r += 8; }
         if (!(x & 0xF0000000)) { x <<= 4;  r += 4; }
@@ -1302,9 +1323,9 @@ value coq_interprete
         
       Instruct (TAIL0INT31) {
 	int r = 0;
-        uint32 x;
+        uint32_t x;
         print_instr("TAIL0INT31");
-	x = (((uint32) accu >> 1) | 0x80000000);
+	x = (((uint32_t) accu >> 1) | 0x80000000);
         if (!(x & 0xFFFF)) { x >>= 16; r += 16; }
         if (!(x & 0x00FF)) { x >>= 8;  r += 8; }
         if (!(x & 0x000F)) { x >>= 4;  r += 4; }
@@ -1354,7 +1375,7 @@ value coq_interprete
         /*accu=accu or accu = (value)((unsigned long)1-accu) if bool
 	               is used for the bits */
         for(i=0; i < 30; i++) {
-	  accu = (value) ((((uint32)accu-1) << 1) | *sp++);
+	  accu = (value) ((((uint32_t)accu-1) << 1) | *sp++);
           /* -1 removes the tag bit, << 1 multiplies the value by 2, 
               | *sp++ pops the last value and add it (no carry involved)
 	      not that it reintroduces a tag bit */
@@ -1374,13 +1395,35 @@ value coq_interprete
         for(i = 30; i >= 0; i--) {
           Field(block, i) = (value)(accu & 3); /* two last bits of the accumulator */
           //Field(block, i) = 3;
-          accu = (value) ((uint32)accu >> 1) | 1;  /* last bit must be a one */
+          accu = (value) ((uint32_t)accu >> 1) | 1;  /* last bit must be a one */
         };
 	accu = block;
         Next;
       }
 
+      Instruct (ORINT31) {
+	/* returns the bitwise or */
+	print_instr("ORINT31");
+        accu =
+	  value_of_uint32((uint32_of_value(accu)) | (uint32_of_value(*sp++)));
+	Next;
+      }
 
+      Instruct (ANDINT31) {
+	/* returns the bitwise and */
+	print_instr("ANDINT31");
+        accu =
+	  value_of_uint32((uint32_of_value(accu)) & (uint32_of_value(*sp++)));
+	Next;
+      }
+
+      Instruct (XORINT31) {
+	/* returns the bitwise xor */
+	print_instr("XORINT31");
+        accu =
+	  value_of_uint32((uint32_of_value(accu)) ^ (uint32_of_value(*sp++)));
+	Next;
+      }
 
       /*  /spiwack   */
 
